@@ -22,22 +22,19 @@
 from __future__ import unicode_literals
 from __future__ import print_function
 
-class PreviousRC1Error(Exception):
+class FedfindError(Exception):
+    """Base class for all fedfind errors."""
+    pass
+
+class PreviousRC1Error(FedfindError):
     """Special exception for trying to get previous_release of a TC1
     or RC1. Makes it easy to spot this case."""
     pass
 
-class WaitError(Exception):
+class WaitError(FedfindError):
     """Error raised when some kind of wait has gone on too long."""
     pass
 
-class RsyncError(Exception):
-    """Exception raised when a necessary rsync operation fails. 'path'
-    is the path we tried to rsync. 'retcode' is the return code.
-    """
-    def __init__(self, path, retcode):
-        self.retcode = retcode
-        self.path = path
-
-    def __str__(self):
-        return "Path: {0} Return code: {1}".format(self.path, self.retcode)
+class DiscoveryError(FedfindError):
+    """Raised when some kind of discovery attempt failed."""
+    pass
