@@ -22,19 +22,16 @@
 from __future__ import unicode_literals
 from __future__ import print_function
 
+import atexit
+import codecs
 import copy
 import datetime
 import json
 import logging
 import os
 import re
-import time
-# Upstream docs recommend using subprocess32 if possible, but make
-# this optional as it's not available in EPEL.
-try:
-    import subprocess32 as subprocess
-except ImportError:
-    import subprocess
+import shutil
+import tempfile
 from six.moves.urllib.parse import urlparse, urlencode
 from six.moves.urllib.request import urlopen, Request
 from six.moves.urllib.error import URLError, HTTPError
@@ -45,6 +42,27 @@ import fedfind.const
 
 logger = logging.getLogger(__name__)
 
+# Cache handling stuff
+CACHEDIR = os.path.join(os.path.expanduser('~'), '.cache', 'fedfind')
+try:
+    if not os.path.exists(CACHEDIR):
+        os.makedirs(CACHEDIR)
+    if not os.access(CACHEDIR, os.W_OK):
+        raise OSError("Write access to {0} denied!".format(CACHEDIR))
+except (IOError, OSError) as err:
+    logger.warning("Cannot use permanent cache! Using temporary cache, valid only for process")
+    logger.warning("Error: %s", err)
+    CACHEDIR = tempfile.mkdtemp()
+    @atexit.register
+    def _clean_cache():
+        """Clean up the temporary cache."""
+        logger.debug("Removing temporary cache %s", CACHEDIR)
+        shutil.rmtree(CACHEDIR, True)
+
+# cache for pkgdb collections JSON, as there are cases where we want
+# to re-use it
+COLLECTIONS = None
+
 def date_check(value, fail_raise=True, out='obj'):
     """Checks whether a value is a date, and returns it if so. Valid
     dates are datetime.date instances or a string or int 'YYYYMMDD'.
@@ -62,8 +80,8 @@ def date_check(value, fail_raise=True, out='obj'):
             dateobj = datetime.datetime.strptime(str(value), '%Y%m%d')
         except ValueError:
             if fail_raise:
-                err = "{0} is not a valid date.".format(value)
-                raise ValueError(err)
+                newerr = "{0} is not a valid date.".format(value)
+                raise ValueError(newerr)
             else:
                 return False
 
@@ -76,40 +94,6 @@ def date_check(value, fail_raise=True, out='obj'):
         else:
             return datestr
 
-def rsync_helper(args, output=False, retry=5):
-    """Runs rsync with specified args. Returns a tuple of the retcode
-    and a string. If output is False, the string will always be empty.
-    If output is True, the string will contain the stdout output. If
-    retry is >0 and rsync fails in a way that looks like some kind of
-    connection error (for now just retcode '5' which is what we get
-    when the server is overloaded), we'll retry that many times, 30
-    seconds apart.
-    """
-    args = list(args)
-    args.insert(0, '/usr/bin/rsync')
-    null = open(os.devnull, 'w')
-    if output:
-        stdout = subprocess.PIPE
-    else:
-        stdout = null
-    # Doesn't use check_output as it's not in Python 2.6
-    process = subprocess.Popen(args, stdout=stdout, stderr=null)
-    out = process.communicate()[0]
-    retcode = process.poll()
-    while retcode == 5 and retry:
-        retry -= 1
-        logger.debug("rsync_helper: server full! Retrying in 30 secs...")
-        time.sleep(30)
-        process = subprocess.Popen(args, stdout=stdout, stderr=null)
-        out = process.communicate()[0]
-        retcode = process.poll()
-    if not out:
-        out = u""
-    # Handle the fact that this is bytes in py3 but str in py2
-    else:
-        out = out.decode()
-    return (retcode, out)
-
 def urlopen_retries(url, tries=5):
     """Simple attempt to open a URL and return the response, with
     retries.
@@ -146,19 +130,8 @@ def url_exists(url):
             return True
         except (ValueError, URLError, HTTPError):
             return False
-    elif 'rsync' in scheme:
-        logger.debug("url_exists: checking %s as rsync", url)
-        retcode = rsync_helper([url], False)[0]
-        # 23 is the 'file not found' return code
-        if retcode == 23:
-            return False
-        elif retcode == 5:
-            raise IOError("url_exists: rsync server busy")
-        elif retcode:
-            raise IOError("Unexpected rsync error {0}!".format(retcode))
-        return True
     else:
-        raise ValueError("Invalid or unhandled URL!")
+        raise ValueError("Invalid or unhandled URL! URL: {0}".format(url))
 
 def get_size(url):
     """Get the size of URL (Content-Length header for HTTP). Currently
@@ -166,7 +139,7 @@ def get_size(url):
     """
     logger.debug("Checking size of %s", url)
     if not url.startswith('http'):
-        raise ValueError("Can only check HTTP URLs for now.")
+        raise ValueError("Can only check HTTP URLs.")
     headers = urlopen_retries(url).info()
     size = headers.get('Content-Length')
     if not size:
@@ -180,15 +153,24 @@ def comma_list(string):
     """
     return string.lower().split(',')
 
+def _get_collections():
+    """Download the collections JSON if we haven't already, and cache
+    it at the module level.
+    """
+    global COLLECTIONS
+    if not COLLECTIONS:
+        logger.debug("collections not cached, downloading...")
+        gotjson = download_json(
+            'https://admin.fedoraproject.org/pkgdb/api/collections/')
+        COLLECTIONS = gotjson['collections']
+    return COLLECTIONS
+
 def get_current_release(branched=False):
     """Find out what the 'current' Fedora release is, according to
     the pkgdb API. If 'branched' is truth-y, will return the Branched
     release number if there is one.
     """
-    logger.debug("get_current_release: finding current release")
-    gotjson = download_json(
-        'https://admin.fedoraproject.org/pkgdb/api/collections/')
-    collections = gotjson['collections']
+    collections = _get_collections()
     if branched:
         rels = [coll['version'] for coll in collections if
                 coll['status'] in ('Active', 'Under Development')]
@@ -198,6 +180,14 @@ def get_current_release(branched=False):
     rels = [int(rel) for rel in rels if rel.isdigit()]
     return max(rels)
 
+def get_current_stables():
+    """Returns a list of the release numbers of all current stable
+    releases (according to pkgdb).
+    """
+    collections = _get_collections()
+    return [int(coll['version']) for coll in collections if
+            coll['status'] == 'Active' and coll['name'] == 'Fedora']
+
 def create_image_dict(path):
     """Synthesize a productmd-style image metadata dict by analyzing
     the image's path. Used for pre-Pungi 4 releases, and Pungi 4
@@ -242,6 +232,10 @@ def create_image_dict(path):
     # and this is just magic.
     if not imgdict['type'] and "disc" in filename:
         imgdict['type'] = "cd"
+    # Pungi config uses the type 'boot' for netinst images, so follow
+    # that
+    if imgdict['type'] == 'netinst':
+        imgdict['type'] = 'boot'
     # Various special handlings for awkward ostree installer images
     # Since 2016-10-10 with up-to-date Pungi the filenames should have
     # -ostree- in them...
@@ -260,6 +254,10 @@ def create_image_dict(path):
     # probably too broad to keep forever, drop when 25 Beta is gone
     if '-workstation-dvd-' in filename:
         imgdict['type'] = "dvd-ostree"
+    # live respins don't have 'live' in the filenames, find them with
+    # the regex
+    if not imgdict['type'] and fedfind.const.RESPINRE.search(path):
+        imgdict['type'] = "live"
 
     # Find subvariant
     imgdict['subvariant'] = ''
@@ -283,19 +281,37 @@ def create_image_dict(path):
     # Try to identify 'generic' media (call them
     # 'Everything' as that's what productmd does)
     if not imgdict['subvariant'] and (imgdict['type'] in
-                                      ('dvd', 'boot', 'netinst', 'cd')):
+                                      ('dvd', 'boot', 'cd')):
         imgdict['subvariant'] = 'Everything'
 
+    # assign disc number
+    # FIXME: can't do disc_count very well with this approach
+    imgdict['disc_number'] = 1
+    match = re.search(r'-disc(\d+)', filename)
+    if match:
+        imgdict['disc_number'] = int(match.group(1))
+
     return imgdict
 
 def pdc_query(path, params=None):
     """Light 'API client', sorta, for PDC. Just handles sending a
     request to a given API path with the appropriate header to get
     a JSON response, parsing the response, and handling multi-page
-    responses. params is a dict of query params.
+    responses. params is a dict of query params. Caches certain
+    responses which are very unlikely ever to change in fedfind's
+    cache directory.
     """
-    results = []
-    baseurl = '/'.join((fedfind.const.PDC_API, path.strip('/')))
+    logger.debug("pdc_query: path is %s", path)
+    # load the query cache
+    cachefile = os.path.join(CACHEDIR, 'pdcquery.json')
+    try:
+        with codecs.open(cachefile, 'r', encoding='utf-8') as cachefh:
+            cache = json.loads(cachefh.read())
+    except (IOError, OSError):
+        # file doesn't exist, etc.
+        cache = {}
+
+    # handle params (as we use them in the cache key)
     if not params:
         params = []
     # turn param dict into tuple list (we need to handle both dicts
@@ -305,6 +321,22 @@ def pdc_query(path, params=None):
         params = list(params.items())
     except AttributeError:
         pass
+    baseurl = '/'.join((fedfind.const.PDC_API, path.strip('/')))
+    # cache key is the URL of the query, but without the page param
+    cachekey = '?'.join((baseurl, urlencode(params)))
+
+    # check cache, return if hit
+    try:
+        ret = cache[cachekey]
+        logger.debug("pdc_query: cache hit!")
+        return ret
+    except KeyError:
+        pass
+
+    # otherwise, actually do the query
+    results = []
+    # we figured out the params earlier, for the cache key, just add
+    # 'page' and re-do the URL
     pgcnt = 1
     params.append(('page', str(pgcnt)))
     url = '?'.join((baseurl, urlencode(params)))
@@ -323,31 +355,84 @@ def pdc_query(path, params=None):
             req = Request(url, headers=headers)
             nxt = resp['next']
         except KeyError:
-            # non-paged query. just return
-            return resp
+            # non-paged query. just take the response
+            results = resp
+            break
+
+    # decide whether to cache results
+    writecache = False
+    # we always cache compose-images results, as if we get a result
+    # it's a 'hit' and will never change (query for non-existent
+    # compose gives a 404)
+    if path.startswith('compose-images/'):
+        logger.debug("pdc_query: caching result for %s", path)
+        cache[cachekey] = results
+        writecache = True
+    # we cache 'composes' results if they're specific and we got a
+    # result
+    if path == 'composes' and results:
+        parkeys = [param[0] for param in params]
+        if 'compose_id' in parkeys or all(key in parkeys for key in ('release', 'compose_label')):
+            logger.debug("pdc_query: caching 'composes' result")
+            cache[cachekey] = results
+            writecache = True
+
+    # write cache if we decided to
+    if writecache:
+        with codecs.open(cachefile, 'w', encoding='utf-8') as cachefh:
+            cachefh.write(json.dumps(cache, separators=(',', ':'), indent=0))
+
     return results
 
 def find_cid(string):
-    """Find a Pungi 4 compose ID in a string."""
-    cidpatt = re.compile(r'Fedora-.+-\d{8,8}(\.\w)?\.\d+')
+    """Find a Fedora compose ID (or a fedfind fake CID) in a string."""
+    # the 'rules' are really pretty lax, see productmd.composeinfo
+    # _validate_id's regex:
+    # r".*\d{8}(\.nightly|\.n|\.ci|\.test|\.t)?(\.\d+)?"
+    # it just has to be a string with 8 digits in it somewhere. We are
+    # stricter, though, for Fedora. We require:
+    # (dist)-(version)-(date)<.type>.(respin)
+    # the type is optional, but all others must be present. The dist
+    # must start with 'Fedora', but can have more text after that (and
+    # can contain dashes, because life is awful; in the rest of the ID
+    # dashes can only be separators).
+    cidpatt = re.compile(r'Fedora.*-[^-]+-\d{8,8}(\.\w)?\.\d+')
     match = cidpatt.search(string)
     if match:
         return match.group(0)
     else:
         return ''
 
-def parse_cid(origcid):
-    """Get release, date, type, and respin values from a Pungi 4
-    compose id.
+def parse_cid(origcid, dist=False):
+    """If dist is false-y, get (release, date, type, respin) values
+    from a Pungi 4 compose id. If dist is truth-y, get (dist, release,
+    date, type, respin) values from a Pungi 4 compose id.
     """
     # Normalize and check cid
     cid = find_cid(origcid)
     if not cid:
         raise ValueError("{0} does not appear to be a valid Pungi 4 Fedora "
                          "compose ID!".format(origcid))
-    release = cid.split('-')[-2].lower()
+    # The dist can have '-' in it, of course, because life is awful
+    # FIXME: This is very stupid and will break on layered products,
+    # which have a lot more - in them. It can theoretically break in
+    # several other cases, but I don't know if Fedora is currently
+    # hitting any of those. I'm currently thinking about how to deal
+    # with this. See:
+    # https://github.com/release-engineering/productmd/issues/77
+    # We may just have to rejig things to require the full compose
+    # metadata and give up on parsing compose IDs at all. That would
+    # be a lot of work for fedfind. A somewhat uglier alternative for
+    # fedfind's purposes would be to give it special knowledge of the
+    # Fedora product shortnames with - in them, or a generic handler
+    # for 'Fedora-(FOO)' or something.
+    (gotdist, release, _) = cid.rsplit('-', 2)
+    release = release.lower()
     (date, typ, respin) = productmd.composeinfo.get_date_type_respin(cid)
-    return (release, date, typ, respin)
+    if dist:
+        return (gotdist, release, date, typ, respin)
+    else:
+        return (release, date, typ, respin)
 
 def cid_from_label(release, label):
     """Get the compose ID for a compose by label. Must also specify
@@ -469,3 +554,94 @@ def identify_image(imgdict, out='tuple', undersub=False, lower=False):
         ret = '-'.join(ret)
 
     return ret
+
+def _get_filelist(flist='fedora', record=None):
+    """Get a (read mode) filehandle for one of the create-filelist
+    generated image lists from the mirrors. flist can be 'fedora',
+    'alt' or 'archive'. record, if set, should be a dict. It will be
+    given 'cache' and 'downfail' keys that are set to True or False
+    depending on whether the cached file was returned, and whether the
+    remote file download failed. This is really just for the use of
+    the tests and shouldn't be considered part of the API.
+    May raise OSError or IOError if cache file cannot be written, or
+    ValueError if imagelist download fails and we don't have a cached
+    copy. There's a bit of hidden magic here where we write the
+    last-modified string from the request header as the first line of
+    the cached file, then always iterate it out before returning the
+    file handle.
+    """
+    if record is not None:
+        record['cache'] = False
+        record['downfail'] = False
+    # Determine the URL
+    if flist not in ('fedora', 'alt', 'archive'):
+        raise ValueError("_get_filelist: must be fedora, archive or alt!")
+    fname = 'imagelist-{0}'.format(flist)
+    url = '/'.join((fedfind.const.HTTPS_DL, flist, fname))
+    logger.debug("_get_filelist: url is %s", url)
+
+    # Determine the cache location
+    cachefile = os.path.join(CACHEDIR, fname)
+    logger.debug("_get_filelist: cachefile is %s", cachefile)
+
+    # Get the cache file handle, if file is present
+    try:
+        cachefh = codecs.open(cachefile, 'r', encoding='utf-8')
+    except (IOError, OSError):
+        cachefh = None
+
+    # Get the response and last modified date
+    try:
+        resp = urlopen_retries(url)
+        modified = resp.headers['last-modified']
+    except ValueError:
+        if cachefh:
+            logger.warning("Could not access image list {0}! Using cached "
+                           "copy. Cache may be outdated!")
+            if record:
+                record['downfail'] = True
+                record['cache'] = True
+            # iterate out the date line
+            next(cachefh)
+            return cachefh
+        else:
+            # We can't do anything useful, just raise the error
+            raise
+
+    # Check if our cache is up-to-date, if present. If it is, go ahead
+    # and return the file handle, with the date line already gone
+    if cachefh:
+        try:
+            cachedate = next(cachefh).strip()
+        except StopIteration:
+            # file is empty
+            cachedate = ''
+        if cachedate == modified:
+            logger.debug("_get_filelist: cache hit, returning")
+            if record:
+                logger.debug("Updating record")
+                record['cache'] = True
+            return cachefh
+
+    # If we got here, either we don't have a cache file at all or the
+    # dates don't match. So we need to re-download it
+    with codecs.open(cachefile + '.tmp', 'w', encoding='utf-8') as cachefh:
+        # Write the modified date line
+        cachefh.write(modified + '\n')
+    with open(cachefile + '.tmp', 'ab') as cachefh:
+        # Now download and write the file, in 1MB chunks at a time
+        while True:
+            chunk = resp.read(1024 * 1024)
+            if chunk:
+                cachefh.write(chunk)
+                cachefh.flush()
+            else:
+                break
+
+    os.rename(cachefile + '.tmp', cachefile)
+
+    # Now open the read-only filehandle, iterate out the date line,
+    # and return
+    cachefh = codecs.open(cachefile, 'r', encoding='utf-8')
+    next(cachefh)
+    return cachefh
