@@ -22,7 +22,10 @@
 from __future__ import unicode_literals
 from __future__ import print_function
 
+import codecs
 import datetime
+import hashlib
+import os
 
 import mock
 import pytest
@@ -36,56 +39,6 @@ import fedfind.helpers
 net = pytest.mark.net
 
 URL = 'https://dl.fedoraproject.org'
-RSYNCURL = 'rsync://dl.fedoraproject.org'
-COLLECTIONS_JSON = {
-  "collections": [
-    {
-      "allow_retire": True,
-      "branchname": "master",
-      "date_created": "2014-05-14 12:36:15",
-      "date_updated": "2016-02-23 22:56:34",
-      "dist_tag": ".fc25",
-      "koji_name": "rawhide",
-      "name": "Fedora",
-      "status": "Under Development",
-      "version": "devel"
-    },
-    {
-      "allow_retire": False,
-      "branchname": "f22",
-      "date_created": "2015-02-10 14:00:01",
-      "date_updated": "2015-02-10 14:00:01",
-      "dist_tag": ".fc22",
-      "koji_name": "f22",
-      "name": "Fedora",
-      "status": "Active",
-      "version": "22"
-    },
-    {
-      "allow_retire": False,
-      "branchname": "f23",
-      "date_created": "2015-07-14 18:13:12",
-      "date_updated": "2015-07-14 18:13:12",
-      "dist_tag": ".fc23",
-      "koji_name": "f23",
-      "name": "Fedora",
-      "status": "Active",
-      "version": "23"
-    },
-    {
-      "allow_retire": True,
-      "branchname": "f24",
-      "date_created": "2016-02-23 22:57:55",
-      "date_updated": "2016-02-25 20:39:53",
-      "dist_tag": ".fc24",
-      "koji_name": "f24",
-      "name": "Fedora",
-      "status": "Under Development",
-      "version": "24"
-    }
-  ],
-  "output": "ok"
-}
 PDC_JSON = {
     "results": ['some', 'results'],
     "next": False
@@ -160,6 +113,7 @@ class FakeResp(object):
     def info(self):
         return {'Content-Length': 100}
 
+@pytest.mark.usefixtures("clean_home")
 class TestHelpers:
     """Tests for the functions in helpers.py."""
     def test_date_check(self):
@@ -194,40 +148,6 @@ class TestHelpers:
         with pytest.raises(ValueError):
             fedfind.helpers.date_check(invalid)
 
-    # we gotta patch the right thing...in fedfind.helpers we import
-    # subprocess32 as subprocess if we can, so follow that here
-    try:
-        import subprocess32
-        subpname = 'subprocess32'
-    except ImportError:
-        subpname = 'subprocess'
-
-    @mock.patch('{0}.Popen.__init__'.format(subpname), return_value=None)
-    @mock.patch('{0}.Popen.communicate'.format(subpname), return_value=(b'foobar', None))
-    @mock.patch('{0}.Popen.poll'.format(subpname), return_value=0)
-    def test_rsync_helper(self, fakepoll, fakecomm, fakesubpinit):
-        assert fedfind.helpers.rsync_helper(['rsync://dl.fedoraproject.org']) == (0, 'foobar')
-        assert fakecomm.call_count == 1
-        assert fakepoll.call_count == 1
-
-    @mock.patch('{0}.Popen.__init__'.format(subpname), return_value=None)
-    @mock.patch('{0}.Popen.communicate'.format(subpname), return_value=(None,None))
-    @mock.patch('{0}.Popen.poll'.format(subpname), return_value=0)
-    def test_rsync_helper_noout(self, fakepoll, fakecomm, fakesubpinit):
-        assert fedfind.helpers.rsync_helper(['rsync://dl.fedoraproject.org']) == (0, '')
-        assert fakecomm.call_count == 1
-        assert fakepoll.call_count == 1
-
-    @mock.patch('{0}.Popen.__init__'.format(subpname), return_value=None)
-    @mock.patch('time.sleep')
-    @mock.patch('{0}.Popen.communicate'.format(subpname), return_value=(None,None))
-    @mock.patch('{0}.Popen.poll'.format(subpname), return_value=5)
-    def test_rsync_helper_busy(self, fakepoll, fakecomm, fakesleep, fakesubpinit):
-        assert fedfind.helpers.rsync_helper(['rsync://dl.fedoraproject.org']) == (5, '')
-        assert fakecomm.call_count == 6
-        assert fakepoll.call_count == 6
-        assert fakesleep.call_count == 5
-
     @mock.patch('fedfind.helpers.urlopen', return_value=True)
     def test_urlopen_retries_good(self, fakeopen):
         assert fedfind.helpers.urlopen_retries(URL) is True
@@ -254,24 +174,6 @@ class TestHelpers:
     def test_url_exists_http_bad(self, fakeopen):
         assert fedfind.helpers.url_exists(URL) is False
 
-    @mock.patch('fedfind.helpers.rsync_helper', return_value=(0, b''))
-    def test_url_exists_rsync_good(self, fakeopen):
-        assert fedfind.helpers.url_exists(RSYNCURL) is True
-
-    @mock.patch('fedfind.helpers.rsync_helper', return_value=(23, b''))
-    def test_url_exists_rsync_bad(self, fakeopen):
-        assert fedfind.helpers.url_exists(RSYNCURL) is False
-
-    @mock.patch('fedfind.helpers.rsync_helper', return_value=(5, b''))
-    def test_url_exists_rsync_busy(self, fakeopen):
-        with pytest.raises(IOError):
-            fedfind.helpers.url_exists(RSYNCURL)
-
-    @mock.patch('fedfind.helpers.rsync_helper', return_value=(6, b''))
-    def test_url_exists_rsync_unknown(self, fakeopen):
-        with pytest.raises(IOError):
-            fedfind.helpers.url_exists(RSYNCURL)
-
     def test_url_exists_invalid(self):
         with pytest.raises(ValueError):
             fedfind.helpers.url_exists('jfohpjsph#^3#@^#')
@@ -286,11 +188,21 @@ class TestHelpers:
         # it splits lists on commas. it ain't rocket science.
         assert fedfind.helpers.comma_list('foo,Bar,moo') == ['foo', 'bar', 'moo']
 
-    @mock.patch('fedfind.helpers.download_json', return_value=COLLECTIONS_JSON)
-    def test_get_current_release(self, fakejson):
+    def test_get_collections(self, json_collections):
+        (mocked, fakedata) = json_collections
+        # download twice to check caching is working
+        colls = fedfind.helpers._get_collections()
+        colls = fedfind.helpers._get_collections()
+        assert colls == fakedata
+        assert mocked.call_count == 1
+
+    def test_get_current_release(self, collections):
         assert fedfind.helpers.get_current_release() == 23
         assert fedfind.helpers.get_current_release(branched=True) == 24
 
+    def test_get_current_stables(self, collections):
+        assert fedfind.helpers.get_current_stables() == [22, 23]
+
     def test_create_image_dict(self):
         # old-style two-week atomic compose
         ret = fedfind.helpers.create_image_dict('Cloud_Atomic/x86_64/iso/Fedora-Cloud_Atomic-x86_64-23-20160313.iso')
@@ -407,6 +319,10 @@ class TestHelpers:
         assert res == 'Fedora-24-20160314.t.0'
         res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-Rawhide-20160314.n.1/compose')
         assert res == 'Fedora-Rawhide-20160314.n.1'
+        res = fedfind.helpers.find_cid('https://foo.com/compose/twoweek/Fedora-Atomic-24-20161004.1/compose')
+        assert res == 'Fedora-Atomic-24-20161004.1'
+        res = fedfind.helpers.find_cid('FedoraRespin-24-20161206.0')
+        assert res == 'FedoraRespin-24-20161206.0'
         # a few 'close but no cigars' just for safety...
         res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-24-20160314.t.u/compose')
         assert res == ''
@@ -418,6 +334,8 @@ class TestHelpers:
     def test_parse_cid(self):
         res = fedfind.helpers.parse_cid('Fedora-24-20160314.n.0')
         assert res == ('24', '20160314', 'nightly', 0)
+        res = fedfind.helpers.parse_cid('Fedora-24-20160314.n.0', dist=True)
+        assert res == ('Fedora', '24', '20160314', 'nightly', 0)
         res = fedfind.helpers.parse_cid('Fedora-24-20160314.0')
         assert res == ('24', '20160314', 'production', 0)
         res = fedfind.helpers.parse_cid('Fedora-24-20160314.t.0')
@@ -427,9 +345,16 @@ class TestHelpers:
         # Two-week Atomic uses a different product name
         res = fedfind.helpers.parse_cid('Fedora-Atomic-24-20160628.0')
         assert res == ('24', '20160628', 'production', 0)
+        res = fedfind.helpers.parse_cid('Fedora-Atomic-24-20160628.0', dist=True)
+        assert res == ('Fedora-Atomic', '24', '20160628', 'production', 0)
         # should raise ValueError for non-Pungi 4-ish CID
         with pytest.raises(ValueError):
             res = fedfind.helpers.parse_cid('23-20160530')
+        # should handle our own fake compose IDs, though
+        res = fedfind.helpers.parse_cid('Fedora-20-19700101.0', dist=True)
+        assert res == ('Fedora', '20', '19700101', 'production', 0)
+        res = fedfind.helpers.parse_cid('FedoraRespin-24-20161206.0', dist=True)
+        assert res == ('FedoraRespin', '24', '20161206', 'production', 0)
 
     @mock.patch('fedfind.helpers.download_json', return_value=PDC_JSON_REAL)
     def test_cid_from_label(self, fakejson):
@@ -459,3 +384,51 @@ class TestHelpers:
         assert fedfind.helpers.identify_image(SERVDVD_DICT, lower=True) == ('server', 'dvd', 'iso')
         assert fedfind.helpers.identify_image(CLOUDRAWXZ_DICT, undersub=False) == ('Cloud_Base', 'raw-xz', 'raw.xz')
         assert fedfind.helpers.identify_image(CLOUDRAWXZ_DICT, undersub=True) == ('Cloud_Base', 'raw_xz', 'raw.xz')
+
+    def test_get_filelist(self, http, clean_home):
+        flists = ('fedora', 'alt', 'archive')
+
+        # inner function for doing the check (re-used to check all
+        # files)
+        def check_file(flist, touch=False):
+            origfile = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'http',
+                                    'pub', flist, 'imagelist-{0}'.format(flist))
+            if touch:
+                os.utime(origfile, None)
+            gotcksum = hashlib.md5()
+            origcksum = hashlib.md5()
+            record = {}
+            with fedfind.helpers._get_filelist(flist=flist, record=record) as testfh:
+                assert testfh.name == os.path.join(clean_home, '.cache', 'fedfind',
+                                                   'imagelist-{0}'.format(flist))
+                for line in testfh:
+                    gotcksum.update(line.encode('utf-8'))
+            with codecs.open(origfile, encoding='utf-8') as origfh:
+                for line in origfh:
+                    origcksum.update(line.encode('utf-8'))
+            assert gotcksum.digest() == origcksum.digest()
+            return record
+
+        for flist in flists:
+            check_file(flist)
+
+        # now re-do the checks and make sure we hit the cache
+        for flist in flists:
+            record = check_file(flist)
+            assert record['cache'] == True
+
+        # now re-do after 'touch'ing the files so the timestamps will
+        # differ; we should re-download now, not hit the cache
+        for flist in flists:
+            record = check_file(flist, touch=True)
+            # we have three 'cache hit' log messages so far
+            assert record['cache'] == False
+
+        # now make the download fail: we should still work using the
+        # cached copy, but warn about it
+        origconst = fedfind.const.HTTPS_DL
+        fedfind.const.HTTPS_DL = 'http://localhost:5001/nonexistent'
+        for flist in flists:
+            check_file(flist)
+        # Reset the constant
+        fedfind.const.HTTPS_DL = origconst
