@@ -22,7 +22,10 @@
 from __future__ import unicode_literals
 from __future__ import print_function
 
+import codecs
 import datetime
+import json
+import os
 import pytest
 
 import mock
@@ -247,6 +250,7 @@ def urlopen_fake_package(url):
     """Stub for urlopen_retries which returns FakeResponsePackages."""
     return FakeResponsePackages(url)
 
+@pytest.mark.usefixtures("clean_home")
 class TestRelease:
     """Tests for release.py."""
 
@@ -270,6 +274,8 @@ class TestRelease:
             (('Rawhide', '', datestr, 2), (fedfind.release.RawhideNightly, 'Rawhide', '', datestr, '2')),
             # Atomic nightly.
             ((24, 'Atomic', datestr, 0), (fedfind.release.AtomicNightly, '24', 'Atomic', datestr, '0')),
+            # Docker nightly.
+            ((25, 'Docker', datestr, 0), (fedfind.release.DockerNightly, '25', 'Docker', datestr, '0')),
             # Wikitcms format Rawhide.
             (('24', 'Rawhide', '{0}.n.0'.format(datestr), None),
              (fedfind.release.RawhideNightly, 'Rawhide', '', datestr, '0')),
@@ -278,6 +284,7 @@ class TestRelease:
             ((23, 'Branched', '', 0), (fedfind.release.BranchedNightly, '23', 'Branched', datestr, '0')),
             (('Rawhide', '', '', 0), (fedfind.release.RawhideNightly, 'Rawhide', '', datestr, '0')),
             ((24, 'Atomic', '', 0), (fedfind.release.AtomicNightly, '24', 'Atomic', datestr, '0')),
+            ((25, 'Docker', '', 0), (fedfind.release.DockerNightly, '25', 'Docker', datestr, '0')),
             # Default, check None is handled.
             ((None, None, None, 0), (fedfind.release.RawhideNightly, 'Rawhide', '', datestr, '0')),
         )
@@ -363,9 +370,9 @@ class TestRelease:
         assert fakecid.call_count == 0
 
     def test_get_release_cid_atomic_new(self):
-        """Test that new (Pungi 4) 2 Week Atomic compose IDs are
-        handled. They have 'Fedora-Atomic' as the product name instead
-        of 'Fedora'.
+        """Test that nightly stable Atomic compose IDs are handled.
+        They have 'Fedora-Atomic' as the product name instead of
+        'Fedora'.
         """
         ret = fedfind.release.get_release(cid='Fedora-Atomic-24-20160628.1')
         assert isinstance(ret, fedfind.release.AtomicNightly)
@@ -373,6 +380,55 @@ class TestRelease:
         assert ret.compose == '20160628'
         assert ret.respin == '1'
 
+    def test_get_release_cid_docker(self):
+        """Test that nightly stable Docker compose IDs are handled.
+        They have 'Fedora-Docker' as the product name instead of
+        'Fedora'.
+        """
+        ret = fedfind.release.get_release(cid='Fedora-Docker-25-20170113.2')
+        assert isinstance(ret, fedfind.release.DockerNightly)
+        assert ret.release == '25'
+        assert ret.compose == '20170113'
+        assert ret.respin == '2'
+
+    def test_get_release_cid_fake_stable(self):
+        """Getting a release based on the fake CID produced for stable
+        releases should return the same release.
+        """
+        rel = fedfind.release.get_release(15)
+        rel2 = fedfind.release.get_release(cid=rel.cid)
+        assert rel2.version == rel.version
+
+    def test_get_release_url_cid(self):
+        """Getting a release based on a correct nightly compose URL
+        should work.
+        """
+        ret = fedfind.release.get_release(
+            url='https://kojipkgs.fedoraproject.org/compose/rawhide/Fedora-Rawhide-20170126.n.0/compose/')
+        assert isinstance(ret, fedfind.release.RawhideNightly)
+        assert ret.release == 'Rawhide'
+        assert ret.compose == '20170126'
+        assert ret.respin == '0'
+
+    def test_get_release_url_cid_unknown(self):
+        """Getting a release based on an unknown URL should return
+        a generic Pungi4Release instance, so long as the URL starts
+        with an acceptable string (to avoid puiterwijk's 'evil URL
+        input' scenario).
+        """
+        ret = fedfind.release.get_release(
+            url='https://kojipkgs.fedoraproject.org/compose/whatthehellisthis/ihavenoidea/compose/')
+        assert isinstance(ret, fedfind.release.Pungi4Release)
+
+    def test_get_release_url_cid_unknown_evil(self):
+        """Getting a release based on an unknown URL that doesn't
+        start with one of our known prefixes should raise a value
+        error.
+        """
+        with pytest.raises(ValueError):
+            ret = fedfind.release.get_release(
+                url='https://foobar.fedoraproject.org/compose/whatthehellisthis/ihavenoidea/compose/')
+
     @mock.patch('fedfind.release.Compose.exists', True)
     def test_get_release_compose(self):
         """A production/candidate compose. This tests the case where
@@ -433,9 +489,15 @@ class TestRelease:
         assert got.compose == datestr
         assert got.release == '23'
 
+        got = fedfind.release.get_release('', 'Docker', '', 0)
+        assert isinstance(got, fedfind.release.DockerNightly)
+        assert got.compose == datestr
+        assert got.release == '23'
+
     @mock.patch('fedfind.release.RawhideNightly', FakeRelease)
     @mock.patch('fedfind.release.BranchedNightly', FakeRelease)
     @mock.patch('fedfind.release.AtomicNightly', FakeRelease)
+    @mock.patch('fedfind.release.DockerNightly', FakeRelease)
     def test_get_release_guess_respin(self):
         got = fedfind.release.get_release(24, 'Branched', '20160314')
         assert isinstance(got, fedfind.release.BranchedNightly)
@@ -446,6 +508,9 @@ class TestRelease:
         got = fedfind.release.get_release(24, 'Atomic', '20160628')
         assert isinstance(got, fedfind.release.AtomicNightly)
         assert got.respin == 3
+        got = fedfind.release.get_release(24, 'Docker', '20160628')
+        assert isinstance(got, fedfind.release.DockerNightly)
+        assert got.respin == 3
 
     @mock.patch('fedfind.release.BranchedNightly', FakeReleaseExists)
     @mock.patch('fedfind.release.AtomicNightly', FakeReleaseExists)
@@ -486,6 +551,56 @@ class TestRelease:
         got = fedfind.release.get_release(25, '', '20161006', 3)
         assert isinstance(got, fedfind.release.BranchedNightly)
 
+    def test_get_release_respinrelease(self, http):
+        """Test get_release for RespinRelease class. We use the fake
+        web server here as RespinRelease has to discover its own
+        release and compose from the image list.
+        """
+        got = fedfind.release.get_release('', 'Respin', '')
+        assert isinstance(got, fedfind.release.RespinRelease)
+        assert got.release == '24'
+        assert got.milestone == 'Respin'
+        assert got.compose == '20161120'
+        # getting a release from the fake compose ID should get the
+        # same release
+        got2 = fedfind.release.get_release(cid=got.cid)
+        assert got.version == got2.version
+        # getting a release from the URL should get the same release
+        got3 = fedfind.release.get_release(
+            url='https://dl.fedoraproject.org/pub/alt/live-respins/')
+        assert got3.version == got2.version
+        # trying to get RespinRelease with a non-matching release
+        # or compose should raise an error
+        with pytest.raises(ValueError):
+            fedfind.release.get_release('25', 'Respin', '')
+        with pytest.raises(ValueError):
+            fedfind.release.get_release('', 'Respin', '20170305')
+        with pytest.raises(ValueError):
+            fedfind.release.get_release('25', 'Respin', '20170305')
+        with pytest.raises(ValueError):
+            fedfind.release.get_release(cid='FedoraRespin-25-20170305.0')
+
+    def test_get_release_cid_passthru(self):
+        """Check compose ID is passed through from get_release to the
+        Release instance in various cases.
+        """
+        datestr = datetime.date.today().strftime('%Y%m%d')
+        rels = (
+            # Branched nightly.
+            (23, 'Branched', datestr, 1, 'Fedora-23-{0}.n.1'.format(datestr)),
+            # Rawhide nightly.
+            ('Rawhide', '', datestr, 2, 'Fedora-Rawhide-{0}.n.2'.format(datestr)),
+            # Atomic nightly.
+            (24, 'Atomic', datestr, 0, 'Fedora-Atomic-24-{0}.0'.format(datestr)),
+            # Docker nightly.
+            (25, 'Docker', datestr, 0, 'Fedora-Docker-25-{0}.0'.format(datestr)),
+        )
+        for (release, milestone, compose, respin, cid) in rels:
+            got = fedfind.release.get_release(release, milestone, compose, respin, cid=cid)
+            if not got._cid:
+                pytest.fail("Testing {0} {1} {2}, CID not passed through!".format(
+                    release, milestone, compose, got.__class__))
+
     @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
     def test_get_package_nvras_fc1(self):
         """FC1: no /Everything, no /source, no split-by-initials."""
@@ -536,6 +651,45 @@ class TestRelease:
             'fakepackage': ''
         }
 
+    def test_all_stable(self, http, pdc):
+        """Test that we get the expected image dicts for all stable
+        releases, using the test HTTP server and home directory and
+        the PDC query mock. Also check there are no missing expected
+        images (as a test for check_expected).
+        """
+        ref = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'allstable.json')
+        with codecs.open(ref, encoding='utf-8') as reffh:
+            expected = json.loads(reffh.read())
+        for relnum in expected:
+            rel = fedfind.release.get_release(relnum)
+            imgs = sorted(rel.all_images, key=lambda x:x['path'])
+            assert expected[relnum] == imgs
+            # we have to force the release to 'exist' or else
+            # check_expected bails
+            rel._exists = True
+            assert not rel.check_expected()
+
+    def test_images_respinrelease(self, http):
+        """Test that we get approximately the expected image dicts for
+        RespinRelease. This depends on the contents of imagelist-alt
+        in the fake http server; if that's updated, this test may have
+        to change. Especially note that the 20161120 respin didn't
+        have an SOAS image.
+        """
+        rel = fedfind.release.RespinRelease()
+        imgs = rel.all_images
+        assert len(imgs) == 7
+        # now ditch source, as we don't care much about it
+        imgs = [img for img in imgs if img['subvariant'] != 'Source']
+        assert len(imgs) == 6
+        assert all(img['type'] == 'live' for img in imgs)
+        # this is a lie, but it's the lie we decided to tell
+        assert all(img['variant'] == 'Spins' for img in imgs)
+        assert all(img['arch'] == 'x86_64' for img in imgs)
+        exp = ['Cinnamon', 'KDE', 'LXDE', 'Mate', 'Workstation', 'Xfce']
+        got = sorted([img['subvariant'] for img in imgs])
+        assert got == exp
+
     # FIXME: Pungi 4 PDC-based get_pkg_nvras method is temporarily
     # disabled due to
     # https://github.com/fedora-infra/pdc-updater/issues/10 so this
